Attribute VB_Name = "ADLCmod"
' 6854 ADLC - minimum required to convince the server that an
' Econet board is installed...
'
' Project: FileStore [E01S emulator]
' Module : ADLCmod
'
' Version: 0.00
' Date   : 5th August 2009
' Started: 20th April 2009
'
' Author : Rick Murray, heyrick1973 -at- yahoo -dot- co -dot- uk
'          http://www.heyrick.co.uk/econet/
'
'
' Licence:
'
'   Copyright 2010 Rick Murray
'
'   Licensed under the EUPL, Version 1.1 ONLY (the "Licence");
'   You may not use this work except in compliance with the Licence.
'
'   You may obtain a copy of the Licence at:
'
'     http://ec.europa.eu/idabc/eupl5
'
'   Unless required by applicable law or agreed to in writing, software
'   distributed under the Licence is distributed on an "AS IS" basis,
'   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
'   implied. See the Licence for the specific language governing
'   permissions and limitations under the Licence.
'
'
'   Rick's note: The EUPL is "compatible" with GPL v2, but *NOT* the more
'                restrictive GPL v3. If you will be involving GPL code
'                in this project, or using some of this code in a GPL
'                project, you should specify the use of ONLY the GPL v2
'                (and not the boilerplate "GPL 2 or any later...").
'

Option Explicit

' This device is complicated - it contains NINE registers which
' are accessed via FOUR addresses, making use of R/!W direction
' and an 'address bit' (control reg. 0, bit 0).
'
'             Addr  AB?  Reg
'    [WRITE]  0  0  -    Control Register #1
'    [WRITE]  0  1  0    Control Register #2
'    [WRITE]  0  1  1    Control Register #3
'    [WRITE]  1  0  -    Transmit FIFO (Frame Continue)
'    [WRITE]  1  1  0    Transmit FIFO (Frame Terminate)
'    [WRITE]  1  1  1    Control Register #4
'
'    [READ]   0  0  -    Status Register #1
'    [READ]   0  1  -    Status Register #2
'    [READ]   1  -  -    Receiver FIFO
'

Private Type ADLCCtrl1Def
  ' Control register #1
  AC As Boolean              ' Address Control
  RIE As Boolean             ' Receiver Interrupt Enable
  TIE As Boolean             ' Transmitter Interrupt Enable
  RDSR As Boolean            ' Receive DMA mode
  TDSR As Boolean            ' Transmit DMA mode
  RxFD As Boolean            ' RX Frame Discontinue
  RxRST As Boolean           ' RX Reset
  TxRST As Boolean           ' TX Reset
End Type

Private Type ADLCCtrl2Def
  ' Control register #2
  PSE As Boolean             ' Prioritised Status Enable
  Byte2Or1 As Boolean        ' 2 byte / 1 byte transfer
  FMIdle As Boolean          ' Flag/Mark Idle
  FCOrTDRA As Boolean        ' Select FC or TRDA status bit
  TxLast As Boolean          ' TX Last data (alt. terminate)
  ClRxSt As Boolean          ' Clear RX status
  ClTxSt As Boolean          ' Clear TX status
  RTS As Boolean             ' Request To Send (is inverted)
End Type

Private Type ADLCCtrl3Def
  ' Control Register #3
  LCF As Boolean             ' Logical Control Field (select)
  ECF As Boolean             ' Extd Control Field (select 16bit)
  AEM As Boolean             ' Auto / Addr Extd Mode
  Idle01Or11 As Boolean      ' Idle condition begins with '0'?
  FDSE As Boolean            ' Flag Detect Status Enable
  Loop As Boolean            ' Loop/Non-loop mode
  Tst As Boolean             ' Test mode bit (is GAP in loop mode)
  DTR As Boolean             ' DTR control (is inverted)
End Type

Private Type ADLCCtrl4Def
  ' Control Register #4
  FFOrF As Boolean           ' Double or Single flags interframe?
  TWLS As Integer            ' (2 bit) TX Word length (5bit-8bit)
  RWLS As Integer            ' (2 bit) RX Word length (5bit-8bit)
  TABT As Boolean            ' TX Abort (sends 8 '1' bits)
  ABTEX As Boolean           ' Extend abort to 16 bits of '1's
  NRZI As Boolean            ' Use NRZI or NRZ coding
End Type

Private Type ADLCStat1Def
  ' Status Register #1
  RDA As Boolean             ' RX Data Available
  S2RQ As Boolean            ' You oughta read Stat. reg #2 now
  LpSt As Boolean            ' Loop status bit (for loop op.)
  FD As Boolean              ' Flag Detected
  CTS As Boolean             ' CTS status
  TxU As Boolean             ' TX Underrun
  TDRAOrFC As Boolean        ' TX Reg Available / Frame Complete
  IRQ As Boolean             ' ADLC IRQ
End Type

Private Type ADLCStat2Def
  ' Status Register #2
  AP As Boolean              ' Address Present
  FVa As Boolean             ' Frame Valid (FV is a VB function!)
  RxI As Boolean             ' Received idle
  RxABT As Boolean           ' Abort received
  FERR As Boolean            ' Frame Error (CRC or short frame)
  DCD As Boolean             ' DCD state
  OVRN As Boolean            ' Receiver Overrun
  RDA As Boolean             ' Receiver Data Available
End Type


Private Type ADLCRegDef
  Control(1 To 4) As Byte    ' the four control regs (byteform)
  Status(1 To 2) As Byte     ' the two status regs (byteform)
  
  TXFIFO(3) As Byte          ' three byte TX FIFO
  RXFIFO(3) As Byte          ' three byte RX FIFO

  ' the flags are also split here into boolean format
  Ctrl1 As ADLCCtrl1Def
  Ctrl2 As ADLCCtrl2Def
  Ctrl3 As ADLCCtrl3Def
  Ctrl4 As ADLCCtrl4Def
  
  Stat1 As ADLCStat1Def
  Stat2 As ADLCStat2Def
End Type

Private ADLC As ADLCRegDef

Public CanIRQ As Boolean
Private InReset As Boolean


Public Sub ADLC_Initialise()
  ' This can double up as a device reset function...

  ' Set device into reset mode (until '0' written to RxRS/TxRS)
  InReset = True
  
  ' Zero everything
  ADLC.Control(1) = 192 ' RxRS & TxRS bits are '1's
  ADLC.Control(2) = 0
  ADLC.Control(3) = 0
  ADLC.Control(4) = 0
  ADLC.Status(1) = 0
  ADLC.Status(2) = 0
  
  ' Split into flags
  ADLC_SplitReg 1
  ADLC_SplitReg 2
  ADLC_SplitReg 3
  ADLC_SplitReg 4
  ADLC_SplitReg -1
  ADLC_SplitReg -2
  
  ' ##TODO## Poke in values so emulator can believe there is
  '          an appropriate ADLC connected and active. In
  '          other words, a working Econet card.
  '
  '
  ' The FileStore MOS startup does this:
  '   ECB5  A9 C1      LDA  #&C1
  '   ECB7  8D 20 FC   STA  &FC20
  '   ECBA  A9 1E      LDA  #&1E
  '   ECBC  8D 23 FC   STA  &FC23
  '   ECBF  9C 21 FC   STZ  &FC21
  '   ECC2  A9 82      LDA  #&82
  '   ECC4  8D 20 FC   STA  &FC20
  '   ECC7  A9 67      LDA  #&67
  '   ECC9  8D 21 FC   STA  &FC21
  '   ECCC  60         RTS
  '
  ' Which in other words is:
  '   Set TxRS/RxRS/AC in CtrlReg1; puts device in reset state.
  '   As AC is set, we write not to the TxFIFO, but to CtrlReg4.
  '   We set the device to:
  '     * Closing flag of 1st frame can be opening flag of 2nd.
  '     * Tx and Rx are 8 bits.
  '     * Use NRZ coding (not NRZI).
  '   As AC is set, we write CtrlReg3. We write a zero which is:
  '     * No Logical Control Field.
  '     * No Ext. Ctrl Field (control field = 8 bits).
  '     * Address field is eight bit.
  '     * Idling condition will be all '1's.
  '     * Flag Detect status is disabled.
  '     * ADLC operates in point-to-point mode, not loop mode.
  '     * Not in Test mode (duh!).
  '     * Set DTR high (is inverted).
  '   We then write back to CtrlReg1 to set:
  '     * Turn address control off (masks out CtrlReg3/CtrlReg4)
  '     * Enable interrupts on data received.
  '     * Tx interrupts remain disabled.
  '     * No DMA stuff (RDSR/TDSR).
  '     * Unset receiver reset, but keep transmitter in reset
  '   Finally, we write to CtrlReg2, to set up:
  '     * Prioritised status mode.
  '     * Single byte data transfer mode.
  '     * Use flag idle mode.
  '     * The FC/TDRA bit is used for TDRA status.
  '     * Clear receiver status bits.
  '     * Clear transmitter status bits.
  '     * RTS is low during data transfer, else high.

  
  ' ##TODO##
  ' In lieu of anything else, we'll fake the DCD bit set
  ADLC.Stat2.DCD = True
End Sub


Public Sub ADLC_Write(ByVal Value As Byte, ByVal Reg As Integer)
  ' This is complicated by the AC bit!
  
  If (ADLC.Ctrl1.AC = False) Then
    ' Address Control NOT set
    Select Case Reg
      Case 0: ' Control Register #1
              ADLC.Control(1) = Value
              ADLC_SplitReg 1
              ' ##TODO## Actions performed on control register 1?
              
              ' We'll check TxRS and RxRS and set overall reset
              ' status on this...
              If ((ADLC.Ctrl1.RxRST = False) And _
                  (ADLC.Ctrl1.TxRST = False)) Then
                ' Take us out of reset mode
                InReset = False
              Else
                InReset = True
              End If
               
      Case 1: ' Control Register #2
              ADLC.Control(2) = Value
              ADLC_SplitReg 2
              ' ##TODO## Actions performed on control register 2?
              
      Case 2: ' Transmit FIFO (frame continue)
              ADLC.TXFIFO(1) = Value
              ' ##TODO## Actions performed on sending a byte?
              ' ##TODO## Prob. shouldn't work if "InReset"!
              
      Case 3: ' Transmit FIFO (frame terminate)
              ADLC.TXFIFO(1) = Value
              ' ##TODO## Actions performed on sending a byte?
              ' set the Ctrl2 TxLast here to fake frame term.?
              ' ##TODO## Prob. shouldn't work if "InReset"!
    End Select
  
   Else
    ' Address Control IS set
    Select Case Reg
      Case 0: ' Control Register #1
              ADLC.Control(1) = Value
              ADLC_SplitReg 1
              ' ##TODO## Actions performed on control register 1?
              
      Case 1: ' Control Register #3
              ADLC.Control(3) = Value
              ADLC_SplitReg 3
              ' ##TODO## Actions performed on control register 3?
              
      Case 2: ' Transmit FIFO (frame continue)
              ADLC.TXFIFO(1) = Value
              ' ##TODO## Actions performed on sending a byte?
              ' ##TODO## Prob. shouldn't work if "InReset"!
              
      Case 3: ' Control Register #4
              ADLC.Control(4) = Value
              ADLC_SplitReg 4
              ' ##TODO## Actions performed on control register 4?
    End Select
  End If
  
  
  HWLog "NET write: " & CStr(Value) & " -> " & CStr(Reg) & _
        " [" & ADLC_LookupReg(Reg) & "  %" & VIA_Binary(Value) & "] " & _
        "CanIRQ = " & Format(CanIRQ, "Yes/No") & ", InReset = " & _
        Format(InReset, "Yes/No")
End Sub


Public Function ADLC_Read(ByVal Reg As Integer)
  Select Case Reg
    Case 0: ' Status Register #1
            ADLC_CombineReg -1
            ADLC_Read = ADLC.Status(1)
            ' ##TODO## Actions performed on reading this reg,
            '          like resetting to IRQ or somesuch.

    Case 1: ' Status Register #2
            ADLC_CombineReg -2
            ADLC_Read = ADLC.Status(2)
            ' ##TODO## Actions performed on reading this reg?

    Case 2, 3: ' Receiver FIFO
            ADLC_Read = ADLC.RXFIFO(1)
            ' ##TODO## Clear the byte etc etc?
            ' ##TODO## Prob. shouldn't work if "InReset"!
  
  End Select
  
  HWLog "NET read : <- " & CStr(Reg) & " [" & _
        ADLC_LookupReg(Reg) & "] CanIRQ = " & _
        Format(CanIRQ, "Yes/No")
End Function


Public Sub ADLC_Poll()
  ' Does nothing
  
End Sub


Private Function ADLC_LookupReg(ByVal Reg As Integer) As String
  Select Case Reg
    Case 0: ADLC_LookupReg = "CR1/SR1"
    Case 1: ADLC_LookupReg = "CR2,3/SR2"
    Case 2: ADLC_LookupReg = "TX/RX FIFOs (FrCont)"
    Case 3: ADLC_LookupReg = "TX/RX FIFOs (FrTerm)"
  End Select
End Function


Private Sub ADLC_SplitReg(ByVal Which As Integer)
  ' Splits the register specified into parts for the boolean
  ' stuff.
  ' Positive values 1 to 4 are Control Registers
  ' Negative values -1 to -2 are Status Registers
  ' Status Registers are not expected to be writeable, it is for
  ' us 'internally'.
  
  If (Which = 1) Then
    ' Control Register #1
    ADLC.Ctrl1.AC = ((ADLC.Control(1) And 1) = 1)
    ADLC.Ctrl1.RIE = ((ADLC.Control(1) And 2) = 2)
    ADLC.Ctrl1.TIE = ((ADLC.Control(1) And 4) = 4)
    ADLC.Ctrl1.RDSR = ((ADLC.Control(1) And 8) = 8)
    ADLC.Ctrl1.TDSR = ((ADLC.Control(1) And 16) = 16)
    ADLC.Ctrl1.RxFD = ((ADLC.Control(1) And 32) = 32)
    ADLC.Ctrl1.RxRST = ((ADLC.Control(1) And 64) = 64)
    ADLC.Ctrl1.TxRST = ((ADLC.Control(1) And 128) = 128)
    
  ElseIf (Which = 2) Then
    ' Control Register #2
    ADLC.Ctrl2.PSE = ((ADLC.Control(2) And 1) = 1)
    ADLC.Ctrl2.Byte2Or1 = ((ADLC.Control(2) And 2) = 2)
    ADLC.Ctrl2.FMIdle = ((ADLC.Control(2) And 4) = 4)
    ADLC.Ctrl2.FCOrTDRA = ((ADLC.Control(2) And 8) = 8)
    ADLC.Ctrl2.TxLast = ((ADLC.Control(2) And 16) = 16)
    ADLC.Ctrl2.ClRxSt = ((ADLC.Control(2) And 32) = 32)
    ADLC.Ctrl2.ClTxSt = ((ADLC.Control(2) And 64) = 64)
    ADLC.Ctrl2.RTS = ((ADLC.Control(2) And 128) = 128)
  
  ElseIf (Which = 3) Then
    ' Control Register #3
    ADLC.Ctrl3.LCF = ((ADLC.Control(3) And 1) = 1)
    ADLC.Ctrl3.ECF = ((ADLC.Control(3) And 2) = 2)
    ADLC.Ctrl3.AEM = ((ADLC.Control(3) And 4) = 4)
    ADLC.Ctrl3.Idle01Or11 = ((ADLC.Control(3) And 8) = 8)
    ADLC.Ctrl3.FDSE = ((ADLC.Control(3) And 16) = 16)
    ADLC.Ctrl3.Loop = ((ADLC.Control(3) And 32) = 32)
    ADLC.Ctrl3.Tst = ((ADLC.Control(3) And 64) = 64)
    ADLC.Ctrl3.DTR = ((ADLC.Control(3) And 128) = 128)

  ElseIf (Which = 4) Then
    ' Control Register #4
    ADLC.Ctrl4.FFOrF = ((ADLC.Control(4) And 1) = 1)
    ADLC.Ctrl4.TWLS = ((ADLC.Control(4) \ 2) And 3)
    ADLC.Ctrl4.RWLS = ((ADLC.Control(4) \ 8) And 3)
    ADLC.Ctrl4.TABT = ((ADLC.Control(4) And 32) = 32)
    ADLC.Ctrl4.ABTEX = ((ADLC.Control(4) And 64) = 64)
    ADLC.Ctrl4.NRZI = ((ADLC.Control(4) And 128) = 128)

  ElseIf (Which = -1) Then
    ' Status Register #1
    ADLC.Stat1.RDA = ((ADLC.Status(1) And 1) = 1)
    ADLC.Stat1.S2RQ = ((ADLC.Status(1) And 2) = 2)
    ADLC.Stat1.LpSt = ((ADLC.Status(1) And 4) = 4)
    ADLC.Stat1.FD = ((ADLC.Status(1) And 8) = 8)
    ADLC.Stat1.CTS = ((ADLC.Status(1) And 16) = 16)
    ADLC.Stat1.TxU = ((ADLC.Status(1) And 32) = 32)
    ADLC.Stat1.TDRAOrFC = ((ADLC.Status(1) And 64) = 64)
    ADLC.Stat1.IRQ = ((ADLC.Status(1) And 128) = 128)

  ElseIf (Which = -2) Then
    ' Status Register #1
    ADLC.Stat2.AP = ((ADLC.Status(2) And 1) = 1)
    ADLC.Stat2.FVa = ((ADLC.Status(2) And 2) = 2)
    ADLC.Stat2.RxI = ((ADLC.Status(2) And 4) = 4)
    ADLC.Stat2.RxABT = ((ADLC.Status(2) And 8) = 8)
    ADLC.Stat2.FERR = ((ADLC.Status(2) And 16) = 16)
    ADLC.Stat2.DCD = ((ADLC.Status(2) And 32) = 32)
    ADLC.Stat2.OVRN = ((ADLC.Status(2) And 64) = 64)
    ADLC.Stat2.RDA = ((ADLC.Status(2) And 128) = 128)
  End If
End Sub


Private Sub ADLC_CombineReg(ByVal Which As Integer)
  ' Rebuilds the register(s) specified from the boolean stuff.
  ' Negative values -1 to -2 are Status Registers
  ' Other registers are not supported, as we will only be reading
  ' the control registers.
  
  Dim StatByte As Byte
  
  StatByte = 0
  
  If (Which = -1) Then
    ' Poke for auto-setting the S2RQ bit
    If ((ADLC.Status(2) And 127) <> 0) Then ADLC.Stat1.S2RQ = True
  
    ' Status Register #1
    If (ADLC.Stat1.RDA) Then StatByte = StatByte Or 1
    If (ADLC.Stat1.S2RQ) Then StatByte = StatByte Or 2
    If (ADLC.Stat1.LpSt) Then StatByte = StatByte Or 4
    If (ADLC.Stat1.FD) Then StatByte = StatByte Or 8
    If (ADLC.Stat1.CTS) Then StatByte = StatByte Or 16
    If (ADLC.Stat1.TxU) Then StatByte = StatByte Or 32
    If (ADLC.Stat1.TDRAOrFC) Then StatByte = StatByte Or 64
    If (ADLC.Stat1.IRQ) Then StatByte = StatByte Or 128
    ADLC.Status(1) = StatByte
    
  ElseIf (Which = -2) Then
    ' Status Register #1
    If (ADLC.Stat2.AP) Then StatByte = StatByte Or 1
    If (ADLC.Stat2.FVa) Then StatByte = StatByte Or 2
    If (ADLC.Stat2.RxI) Then StatByte = StatByte Or 4
    If (ADLC.Stat2.RxABT) Then StatByte = StatByte Or 8
    If (ADLC.Stat2.FERR) Then StatByte = StatByte Or 16
    If (ADLC.Stat2.DCD) Then StatByte = StatByte Or 32
    If (ADLC.Stat2.OVRN) Then StatByte = StatByte Or 64
    If (ADLC.Stat2.RDA) Then StatByte = StatByte Or 128
    ADLC.Status(2) = StatByte
  End If
End Sub

