Attribute VB_Name = "FDCmod"
' 2793 FDC emulation
'
' Project: FileStore [E01S emulator]
' Module : FDCmod
'
' Version: 0.00
' Date   : 8th August 2009
' Started: 20th April 2009
'
' Author : Rick Murray, heyrick1973 -at- yahoo -dot- co -dot- uk
'          http://www.heyrick.co.uk/econet/
'
'
' Licence:
'
'   Copyright 2010 Rick Murray
'
'   Licensed under the EUPL, Version 1.1 ONLY (the "Licence");
'   You may not use this work except in compliance with the Licence.
'
'   You may obtain a copy of the Licence at:
'
'     http://ec.europa.eu/idabc/eupl5
'
'   Unless required by applicable law or agreed to in writing, software
'   distributed under the Licence is distributed on an "AS IS" basis,
'   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
'   implied. See the Licence for the specific language governing
'   permissions and limitations under the Licence.
'
'
'   Rick's note: The EUPL is "compatible" with GPL v2, but *NOT* the more
'                restrictive GPL v3. If you will be involving GPL code
'                in this project, or using some of this code in a GPL
'                project, you should specify the use of ONLY the GPL v2
'                (and not the boilerplate "GPL 2 or any later...").
'
'
' The FDC2793 seems to be a rather simple (in comparison) device that is,
' astonishingly, incapable of even selecting multiple drives and/or
' determining which side of the disc to write to; external logic adds this
' functionality. This FDC is very similar to the 1770.
'
' For reference, we are running at 250kHz in MFM mode. The disc format is
' specific to the FileStore, but would appear to be based upon the BBC
' Master ADFS 'L' format (~640Kb per disc, double sided double density 3.5",
' 256 bytes/sector, MFM and 16 sectors/track for 80 tracks)...
'
'
'              #################################################
'              #                                               #
'              #  THIS IS NOT A COMPLETE EMULATION OF THE FDC  #
'              #  only what is nescessary for the FileStore... #
'              #                                               #
'              #################################################
'
'
' IMPORTANT: Note that the Latch.FDCxxx values are INVERTED to
'            be logically correct. This is because on the
'            actual chip they are active low...

Option Explicit



Private DriveFour As String     ' Name of left disc image file, or "" if empty.
Private DriveFive As String     ' Name of right disc image file, or "" if empty.
Private LeftImage As Integer    ' File handle to left :4 image.
Private RightImage As Integer   ' File handle to right :5 image.
Private CommandReg As Integer   ' Command register
Private TrackNum As Integer     ' FDC track number
Private SectorNum As Integer    ' FDC sector number
Private DataReg As Integer      ' FDC data register
Private LastStatus As Integer   ' Last value for status register
Private StepOut As Boolean      ' Stepping in or stepping out?
Private UpdateTrack As Boolean  ' Update track register when command done?
Private FDCIRQ As Boolean       ' An active IRQ?
Private FDCCommand As Integer   ' Current command, parsed.
Private FDCTimeout As Long      ' Waiting to time out an activity?
Private IsBusy As Boolean       ' Is FDC 'busy'?

' Information specific to each drive
Private Type DiscParamsDef
  SpunUp As Boolean             ' Is the drive spun up?
  CurrTrack As Integer          ' Current track (used to calc image offset)
  Timeout As Long               ' Timeout for spindown
End Type
Private DiscParams(1 To 2) As DiscParamsDef


Private Const CMD_RESTORE        As Integer = 1
Private Const CMD_SEEK           As Integer = 2
Private Const CMD_STEP           As Integer = 3
Private Const CMD_STEPIN         As Integer = 4
Private Const CMD_STEPOUT        As Integer = 5
Private Const CMD_READSECTOR     As Integer = 6
Private Const CMD_WRITESECTOR    As Integer = 7
Private Const CMD_READADDRESS    As Integer = 8
Private Const CMD_READTRACK      As Integer = 9
Private Const CMD_WRITETRACK     As Integer = 10
Private Const CMD_FORCEINTERRUPT As Integer = 11


' The AFS0 disc topology is the same as ADFS-L, which is:
'   Sectors . . . . . . . . 256
'   Sectors per track . . . 16
'   Heads . . . . . . . . . 2
'   Tracks  . . . . . . . . 80
'   Disc size is therefore  655360 bytes (640K), or 256 * 16 * 2 * 80.

Private Const SECTORSIZE         As Long = 256&
Private Const TRACKSIZE          As Long = 8192&



' One of the first things you learn when writing an emulator is that just because
' you can read in a 640K disc image... doesn't mean you can spit it to the emulated
' hardware as-and-when. Things in the real world take time, time upon which we rely
' for a bunch of behind-the-scenes stuff [okay, we have a byte from the disc, well,
' the firmware must now *do* something with it...]
'
' The following delays are measured in CPU CYCLES. Most of these are just "estimates"
' and can probably be tuned for better operation.
Private Const DELAY_SPINUP As Long = 2000000  ' Drive spinup time - a second
Private Const DELAY_STEP As Long = 3000       ' Track step time
Private Const DELAY_SPIN As Long = 200000     ' Rotate time - 1/10 sec
Private Const DELAY_BYTE As Long = 120        ' About 60s between each byte read
Private Const DELAY_TIMEOUT As Long = 3000000 ' Activity timeout, 1 1/2 seconds.


Public Sub FDC_Initialise()
  ' At this point, the strings <DriveFour> and <DriveFive> should have been
  ' either set to a full path to an image file, or blank (for no disc).
  '
  ' The LEFT drive is known to us as drive 1 (DiscParams(1)), and to the
  ' FileStore as :4. Accordingly, the RIGHT drive is 2 to us, or :5 to FS.

  ' Blank status
  CommandReg = 0
  TrackNum = 0
  SectorNum = 0
  DataReg = 0
  LastStatus = 0
  StepOut = False
  UpdateTrack = False
  FDCIRQ = False
  FDCCommand = 0
  FDCTimeout = 0
  IsBusy = False
  
  ' Blank disc parameters
  DiscParams(1).CurrTrack = 0
  DiscParams(1).SpunUp = False
  DiscParams(1).Timeout = 0
  DiscParams(2).CurrTrack = 0
  DiscParams(2).SpunUp = False
  DiscParams(2).Timeout = 0
  
  ' Files?
  If (LeftImage <> 0) Then Close #LeftImage
  If (RightImage <> 0) Then Close #RightImage
  If (Len(DriveFour) <> 0) Then
    LeftImage = FreeFile
    Open DriveFour For Binary Access Read As #LeftImage ' #### READ ONLY ### ##TODO##
  End If
  If (Len(DriveFive) <> 0) Then
    RightImage = FreeFile
    Open DriveFive For Binary Access Read As #RightImage ' #### READ ONLY ### ##TODO##
  End If
  
  ' Now do a 'restore' to set us up in a known state; datasheet
  ' does not appear to specify what happens at power-up, so...
  CommandReg = 8
  FDC_Command
End Sub


Public Sub FDC_Write(ByVal Value As Byte, ByVal Reg As Integer)
  ' Writing to an FDC register
  '
  '   Reg      Purpose   Write behaviour
  '   =================================================================
  '    0       Command   Specify an action to perform, and initiate it.
  '    1       Track     Specify a track, no further action.
  '    2       Sector    Specify a sector, no further action.
  '    3       Data      Writes data; in response to command or DRQ.
  '
  ' ** WE ARE CURRENTLY HARDWIRED TO OPERATE IN A READ-ONLY MODE (write protect
  '    marked as set) - WRITING WILL BE LOOKED AT *AFTER* READ VERIFIED OK.
  '
  
  HWLog "FDC write : " & CStr(Value) & " -> " & CStr(Reg) & _
        " [" & FDC_LookupReg(Reg) & "]"
  
  Select Case Reg
    Case 0: ' Command register
            ' Usually perform an action and then throw an interrupt.
            CommandReg = Value
            FDC_Command
    
    Case 1: ' Track register; only to be written if FDC not 'busy'.
            If (IsBusy = False) Then TrackNum = Value
            
    Case 2: ' Sector register; only to be written if FDC not 'busy'.
            If (IsBusy = False) Then SectorNum = Value
            
    Case 3: ' Data register
            DataReg = Value
  End Select
End Sub


Public Function FDC_Read(ByVal Reg As Integer)
  ' Writing to an FDC register
  '
  '   Reg      Purpose   Read behaviour
  '   =================================
  '    0       Status    Read FDC status, bits depend upon last command.
  '    1       Track     Read track number.
  '    2       Sector    Read sector number.
  '    3       Data      Read data.
  '
  
  Select Case Reg
    Case 0: ' Status
            FDCIRQ = False ' Clear interrupts
            FDC_Read = LastStatus
    
    Case 1: ' Track register
            FDC_Read = TrackNum
            
    Case 2: ' Sector register
            FDC_Read = SectorNum
            
    Case 3: ' Data register
            FDCIRQ = False
            FDC_Read = DataReg
  End Select

  HWLog "FDC read : " & CStr(FDC_Read) & " <- " & CStr(Reg) & _
        " [" & FDC_LookupReg(Reg) & "]"

End Function


Public Sub FDC_ChangeState()
  ' State changed (drive, side, density...) from programmable latch.
  ' Nothing really to do here, except report it for debug.
  
  Dim OutStr As String
  
  ' Remember, FDCxxx signals are inverted to make sense. :-)
  OutStr = "FDC change" & _
    ": Test=" & Format(Latch.FDCTest, "Yes/No") & _
    ", MRst=" & Format(Latch.FDCMasteReset, "Yes/No") & _
    ", DDen=" & Format(Latch.FDCDiscDensity, "Yes/No") & _
    "; FDC1=" & Format(Latch.Floppy1, "Yes/No") & _
    ", FDC2=" & Format(Latch.Floppy2, "Yes/No") & _
    ", Side="
  If (Latch.FloppySide = False) Then
    OutStr = OutStr & "0"
  Else
    OutStr = OutStr & "1"
  End If
    
  HWLog OutStr
End Sub


Public Sub FDC_Poll(ByVal HowLong As Long)
  ' Regular "background" disc activity
  Dim ThisDrive As Integer
  
  ' Outstanding IRQ?
  ' (if we're handling an/this IRQ, then we'll just
  '  bounce back here as the I flag will be set.)
  If (FDCIRQ = True) Then CPU_Vector IRQVEC
  
  ' Disc activity spindown timers?
  If (DiscParams(1).SpunUp = True) Then
    DiscParams(1).Timeout = DiscParams(1).Timeout - HowLong
    If (DiscParams(1).Timeout < 0) Then
      DiscParams(1).SpunUp = False
    End If
    FDC_LoadUnloadHead
  End If
  If (DiscParams(2).SpunUp = True) Then
    DiscParams(2).Timeout = DiscParams(2).Timeout - HowLong
    If (DiscParams(2).Timeout < 0) Then
      DiscParams(2).SpunUp = False
    End If
    FDC_LoadUnloadHead
  End If
  
  ' Nothing to do?
  If (IsBusy = False) Then Exit Sub
  
  ' Subtract timer
  If (IsBusy = True) Then FDCTimeout = FDCTimeout - HowLong
  
  ' Expired?
  If (FDCTimeout < 0) Then
    ' Let's deal with the command!
    
    ' Flag no seek/CRC error unless it goes bang.
    FDC_ClearStatusBit 3 ' CRC
    FDC_ClearStatusBit 4 ' Seek
    
    ' ### HARDWIRED FOR NOW ###                                 ##TODO##
    FDC_SetStatusBit 6 ' Disc is write protected
    
    ' Set a variable to make easy access to the current drive.
    ThisDrive = IIf((Latch.Floppy1 = True), 1, 2)
    
    ' Deal first with the Type I commands
    If (FDCCommand <= CMD_STEPOUT) Then
      Select Case FDCCommand
        Case CMD_RESTORE:
          ' Go to track zero
          DiscParams(ThisDrive).CurrTrack = 0
          TrackNum = 0
            
        Case CMD_SEEK:
          ' Go to the track specified in the Data register
          DiscParams(ThisDrive).CurrTrack = DataReg
          ' datasheet (p13) doesn't say if the track register is updated, as the
          ' bit that would be the track register update is 1, we'll update it...
          TrackNum = DataReg
         
        Case CMD_STEP:
          ' Step
          If (StepOut = True) Then
            DiscParams(ThisDrive).CurrTrack = DiscParams(ThisDrive).CurrTrack + 1
          Else
            DiscParams(ThisDrive).CurrTrack = DiscParams(ThisDrive).CurrTrack - 1
          End If
          If (UpdateTrack = True) Then TrackNum = DiscParams(ThisDrive).CurrTrack
      
        Case CMD_STEPIN:
          ' Step in (towards high track, which to me would seem like a step OUT!)
          DiscParams(ThisDrive).CurrTrack = DiscParams(ThisDrive).CurrTrack + 1
          If (UpdateTrack = True) Then TrackNum = DiscParams(ThisDrive).CurrTrack
            
        Case CMD_STEPOUT:
          ' Step in (towards track zero, which to me would seem like a step IN!)
          DiscParams(ThisDrive).CurrTrack = DiscParams(ThisDrive).CurrTrack - 1
          If (UpdateTrack = True) Then TrackNum = DiscParams(ThisDrive).CurrTrack
      End Select
      
      ' Flag track zero?
      If (DiscParams(ThisDrive).CurrTrack = 0) Then
        FDC_SetStatusBit 2 ' is track zero
      Else
        FDC_ClearStatusBit 2 ' is not track zero
      End If
      
      ' We're not busy, and we're not engaging the head any more
      FDC_ClearStatusBit 0 ' not Busy
      FDC_ClearStatusBit 5 ' not Head Loaded
      
      ' Drive is ready, whoo-hoo... or something. :-)
      FDC_ClearStatusBit 7 ' not Not Ready
      
      ' FDC no longer busy
      IsBusy = False
      
      ' Set up the timeout for spindown
      DiscParams(ThisDrive).SpunUp = True ' ought to be True anyway...
      DiscParams(ThisDrive).Timeout = DELAY_TIMEOUT
      
      ' Now flag an interrupt (FDC is on NMIs)
      CPU_Vector NMIVEC
      
      
    ' else...
      
    End If ' ... more to follow
    
    
  End If
  
  
End Sub


Private Function FDC_LookupReg(ByVal Reg As Integer) As String
  ' This just translates register to description for debug report.
  Select Case Reg
    Case 0: FDC_LookupReg = "Status/Command"
    Case 1: FDC_LookupReg = "Track"
    Case 2: FDC_LookupReg = "Sector"
    Case 3: FDC_LookupReg = "Data"
  End Select
End Function


Private Sub FDC_Command()
  ' Deal with a command written to the command register.
  '
  ' There are eleven commands, of four different types. The upper three bits
  ' of the byte mostly distinguish the command, while the remainder of the
  ' bits pass flags.
  '
  '   Command          Type      Bits (left=msb)
  '   ===============================================
  '   Restore          I      0  0  0  0  h  V  r1 r0
  '   Seek             I      0  0  0  1  h  V  r1 r0
  '   Step             I      0  0  1  T  h  V  r1 r0
  '   Step in          I      0  1  0  T  h  V  r1 r0
  '   Step out         I      0  1  1  T  h  V  r1 r0
  '   Read sector      II     1  0  0  m  S  E  C  0
  '   Write sector     II     1  0  1  m  S  E  C  a0
  '   Read address     III    1  1  0  0  0  E  0  0
  '   Read track       III    1  1  1  0  0  E  0  0
  '   Write track      III    1  1  1  1  0  E  0  0
  '   Force Interrupt  IV     1  1  0  1  I3 I2 I1 I0
  '
  '   Drive timeout    n/a    1  1  1  1  1  1  1  1
  '
  ' Where:
  '   h  = Head load flag (1=load head) [means motor 'active']
  '   V  = Track number verify [ignored]
  '   r# = Stepping motor rate [ignored]
  '   T  = Track update (1=update track register)
  '   m  = Multiple record flag
  '   S  = Side compare flag (0=side0, 1=side1)
  '   E  = Extra (15ms) delay flag [ignored]
  '   C  = Side compare flag (0=disable, 1=enable)
  '   a0 = Data address mark
  '   I# = Interrupt flags
  '
  ' The pseudocommand 255 is used by us in order to flag when the FDC has
  ' finished with a drive, but the disc is still spinning, so we can better
  ' emulate the motor stopping as it would in real life. This is not a
  ' command that would be received by the FDC, it would, if anything, be
  ' mistaken for Write track...
  
  Dim Upperbits As Integer, UpdateTrack As Boolean, ThisDrive As Integer
  
  
  ' ### HARDWIRED FOR NOW ###                                 ##TODO##
  FDC_SetStatusBit 6 ' Disc is write protected
  
  ' Clear any outstanding interrupts
  FDCIRQ = False
  
  ' Sort out the upper three bits
  Upperbits = (CommandReg And &HE0)
  
  ' Now figure out what to do
  Select Case Upperbits
    Case Is < &H80:
      ' Type I commands - Restore, Seek, Step, Step-in, Step-out
      
      ' For reference - we only RECORD the desired track. The actual offset is
      ' generated at read/write time as a function of track+side+sector...
      '
      ' I *think* my disc images are interleaved, but I'll need to fire up the
      ' Quattro^WRiscPC to confirm this in a sector editor, basically 'cos the
      ' RedSquirrel emulator's FDC code is too lame to cope correctly with 256
      ' byte/sector images. I guess you'd need VirtualAcorn for that...?
      
      ' Set status bits for all type Is - we're Busy, there's no CRC or Seek error.
      FDC_SetStatusBit 0    ' Busy
      FDC_ClearStatusBit 3  ' no CRC error
      FDC_ClearStatusBit 4  ' no Seek error
      
      If (Upperbits = 0) Then
        If ((CommandReg And &H10) = 0) Then
          ' Restore (seeks to track zero)
          HWLog "FDC: Restore"
          FDCCommand = CMD_RESTORE
                            
        Else
          ' Seek
          HWLog "FDC: Seek"
          FDCCommand = CMD_SEEK
        End If
      
      ElseIf (Upperbits = &H20) Then
        ' Step
        HWLog "FDC: Step"
        FDCCommand = CMD_STEP
        UpdateTrack = (CommandReg And &H10)
            
      ElseIf (Upperbits = &H40) Then
        ' Step in
        HWLog "FDC: Step in"
        FDCCommand = CMD_STEPIN
        StepOut = False
        UpdateTrack = (CommandReg And &H10)
      
      ElseIf (Upperbits = &H60) Then
        ' Step out
        HWLog "FDC: Step out"
        FDCCommand = CMD_STEPOUT
        StepOut = True
        UpdateTrack = (CommandReg And &H10)
      End If
      
      ' We take a few liberties here - we always spin up the drive (head loaded)
      ' and we ignore the step rate. We do, however, log this to note if we ought
      ' to support different functionality later...
      HWLog "FDC command type I, params = " & CStr(CommandReg And 15)
       
      ThisDrive = IIf((Latch.Floppy1 = True), 1, 2)
      If (DiscParams(ThisDrive).SpunUp = False) Then
        ' Spin up
        FDCTimeout = DELAY_SPINUP ' pass some time
        IsBusy = True
        FDC_SetStatusBit 5 ' Head loaded
        DiscParams(ThisDrive).Timeout = DELAY_TIMEOUT
        FDC_LoadUnloadHead
      Else
        ' Just timeout a spin
        FDCTimeout = DELAY_SPIN
        IsBusy = True
      End If
      
      FDC_SetStatusBit 7 ' Not ready - awaiting timeout and interrupt
      
      
    Case &H80:
      ' Type II command - Read sector
      HWLog "FDC: Read sector - TO DO"
      '  ==do it==


    Case &HA0:
      ' Type II command - Write sector
      HWLog "FDC: Write sector - TO DO"
      '  ==do it==
    
    
    Case &HC0:
      ' Type III/IV command - Read address or Force Interrupt
      If ((CommandReg And &H10) = 0) Then
        ' Read address
        HWLog "FDC: Read address - TO DO"
        '  ==do it==
        
      Else
        ' Force interrupt
        HWLog "FDC: Force interrupt - TO DO"
        '  ==do it==
        
        ' hack to do IRQkill function
        If ((CommandReg And 15) = 0) Then FDCIRQ = False
      End If
        
        
    Case &HF0:
      ' Type III commands - Read track, Write track
      If ((CommandReg And &H10) = 0) Then
        ' Read track
        HWLog "FDC: Read track - TO DO"
        '  ==do it==
      
      Else
        ' Write track
        HWLog "FDC: Write track - TO DO"
        '  ==do it==
      End If
      
  End Select
End Sub


Private Sub FDC_LoadUnloadHead()
  ' From SpunUp value, set the UI drive LEDs appropriately
  frmMain.lblLeftInd.Visible = DiscParams(1).SpunUp
  frmMain.lblCLeftInd.Visible = DiscParams(1).SpunUp
  frmMain.lblRightInd.Visible = DiscParams(2).SpunUp
  frmMain.lblCRightInd.Visible = DiscParams(2).SpunUp

  frmMain.lblLeftInd.Refresh
  frmMain.lblCLeftInd.Refresh
  frmMain.lblRightInd.Refresh
  frmMain.lblCRightInd.Refresh
End Sub


Private Sub FDC_SetStatusBit(ByVal Which As Integer)
  ' It would be a nice if we could "|= 1 << Which", but that's too
  ' complicated for poor ittle-bittle Visual Bas, so we cheat
  ' with the power-of.
  ' And how many cycles does this use, over a plain shift? Can
  ' be done more or less on-the-fly on the ARM. ;-)
  
  LastStatus = LastStatus Or (2 ^ Which)
End Sub


Private Sub FDC_ClearStatusBit(ByVal Which As Integer)
  ' This ought to be "&= ~ (1 << Which)"...
  
  LastStatus = LastStatus And (Not (2 ^ Which))
End Sub

