Attribute VB_Name = "NVRAMMod"
' Incomplete emulation of a 146818 RTC/NVRAM.
' Note that WRITING the RTC is silently ignored; and that
' READING it will return values from the host computer.
' This means you can never set the RTC to an arbitrary date,
' and furthermore the RTC will always be in sync with reality.
'
' Project: FileStore [E01S emulator]
' Module : NVRAMmod
'
' Version: 0.00
' Date   : 5th November 2009
' Started: 19th April 2009
'
' Author : Rick Murray, heyrick1973 -at- yahoo -dot- co -dot- uk
'          http://www.heyrick.co.uk/econet/
'
'
' Licence:
'
'   Copyright 2010 Rick Murray
'
'   Licensed under the EUPL, Version 1.1 ONLY (the "Licence");
'   You may not use this work except in compliance with the Licence.
'
'   You may obtain a copy of the Licence at:
'
'     http://ec.europa.eu/idabc/eupl5
'
'   Unless required by applicable law or agreed to in writing, software
'   distributed under the Licence is distributed on an "AS IS" basis,
'   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
'   implied. See the Licence for the specific language governing
'   permissions and limitations under the Licence.
'
'
'   Rick's note: The EUPL is "compatible" with GPL v2, but *NOT* the more
'                restrictive GPL v3. If you will be involving GPL code
'                in this project, or using some of this code in a GPL
'                project, you should specify the use of ONLY the GPL v2
'                (and not the boilerplate "GPL 2 or any later...").
'

Option Explicit

' ##TODO##
' How much extra is actually required for the FileStore?
' We should probably do better if "SET" is active? We don't
' really want to run the risk of firing off a time IRQ when
' we think we're updating the time!
'
' Also, does the FileStore ever use other values for the flags?
' If not, nothing much else to do really. [I hope!]
'
' It seems kinda hairy, but the server's internal ticker is a
' once-per-second based upon the time-updated interrupt; and the
' MOS cs ticker is simply +100 each time. Yikes!


Private Type NVRAM_RTCDef
  Secs As Byte               ' &00 = Seconds
  SecsAlm As Byte            ' &01 = Seconds (Alarm)
  Mins As Byte               ' &02 = Minutes
  MinsAlm As Byte            ' &03 = Minutes (Alarm)
  Hrs As Byte                ' &04 = Hours (12h or 24h)
  HrsAlm As Byte             ' &05 = Hours (Alarm; 12h or 24h)
  DotW As Byte               ' &06 = Day of the Week (1=Sun)
  DotM As Byte               ' &07 = Day (1-31)
  Mnts As Byte               ' &08 = Month (1-12)
  Yrs As Byte                ' &09 = Year

  ' It looks like the FileServer stores the year value as
  ' 0 = 1981; and the NVRAM can hold values up to 99, so
  ' the effective range is 1981-2081? I could have sworn
  ' I got the date '1900' back from the server once? Maybe
  ' this is a quirk in NetFS? I will have to investigate...
End Type


Private Type NVRAM_RegDef
  RegA As Byte               ' &0A = (internal) Register A
  RegB As Byte               ' &0B = (internal) Register B
  RegC As Byte               ' &0C = (internal) Register C
  RegD As Byte               ' &0D = (internal) Register D
End Type

Private Type NVRAM_FlagsDef
  ' Refer to a datasheet for more information.
  
  ' RegisterA
  DivRS As Integer           ' Divider chain bits RS0-RS3
  DivDV As Integer           ' Divider chain bits DV0-DV2
  UIP As Boolean             ' Update In Progress
  
  ' RegisterB
  fSET As Boolean            ' SET bit (is fSET as VB has 'Set')
  PIE As Boolean             ' Periodic Update Enable
  AIE As Boolean             ' Alarm Interrupt Enable
  UIE As Boolean             ' Update-ended Interrupt Enable
  SQWE As Boolean            ' SQuare Wave Enable
  DM As Boolean              ' Data Mode (1 = bin, 0 = BCD)
  f2412 As Boolean           ' 24/12 (1 = 24h, 0 = 12h)
  DSE As Boolean             ' Daylight Save Enable (unsupported?)

  ' Register C
  IRQF As Boolean            ' Interrupt ReQuest Flag
  PF As Boolean              ' Periodic interrupt Flag
  AF As Boolean              ' Alarm interrupt Flag
  UF As Boolean              ' Update-ended interrupt Flag
  ' lower four bits are zero
  
  ' Register D
  VRT As Boolean             ' Valid Ram and Time
  ' lower seven bits are zero
End Type

Private Type NVRAM_RAMDef
  RAM(14 To 63) As Byte
End Type

Private Type NVRAMDef
  RTC As NVRAM_RTCDef
  Reg As NVRAM_RegDef
  Flags As NVRAM_FlagsDef
  RAM As NVRAM_RAMDef
End Type
Private NVRAM As NVRAMDef

Private NVRAMAddr As Integer
Private NVRAMCyc As Long


Public Sub NVRAM_Initialise()
  ' Sets up the NVRAM, loads stored data into it
  
  Dim FP As Integer, Recurse As Integer
  
  If (Len(Dir(App.Path & "\nvram.dat")) > 0) Then
    FP = FreeFile
    Open App.Path & "\nvram.dat" For Binary Access Read As FP
    Get #FP, , NVRAM.RTC
    Get #FP, , NVRAM.Reg
    Get #FP, , NVRAM.RAM
    Close #FP
  End If
  
  ' If no file, just don't initialise, it'll be up to the
  ' FileServer code to set it up correctly. [of course, it'll
  ' probably be zeroes, this is VB not C! :-)]
  
  If (SetupInfo.StartNVRAMClear = True) Then
    ' user requested NVRAM to be cleared, force it
    ' [this will *NOT* affect the RTC due to how it operates]
    
    NVRAM.RTC.SecsAlm = 0
    NVRAM.RTC.MinsAlm = 0
    NVRAM.RTC.HrsAlm = 0
    
    NVRAM.Reg.RegA = 0
    NVRAM.Reg.RegB = 0
    NVRAM.Reg.RegC = 0
    NVRAM.Reg.RegD = 0
    
    For Recurse = 14 To 63
      NVRAM.RAM.RAM(Recurse) = 0
    Next
  End If
  
  ' Decode the flags from the internal registers
  NVRAM_RegDecode
  
  ' Now fake a reset of the part (set a known state)
  NVRAM_Reset
  
  ' Mark NO IRQ on updates
  NVRAMCyc = 0
End Sub


Public Sub NVRAM_Reset()
  ' fakes a reset of the NVRAM
  
  ' As defined in RS datasheet 232-2605 p9
  NVRAM.Flags.PIE = False
  NVRAM.Flags.AIE = False
  NVRAM.Flags.UIE = False
  ' IRQF is set False by PF=AF=UF=False
  NVRAM.Flags.UF = False
  NVRAM.Flags.PF = False
  NVRAM.Flags.AF = False
  NVRAM.Flags.SQWE = False
  NVRAM_RegEncode
End Sub


Public Sub NVRAM_Save()
  ' Saves the current NVRAM settings
  
  Dim FP As Integer, Recurse As Integer
  
  If (Len(Dir(App.Path & "\nvram.dat")) > 0) Then
    Kill App.Path & "\nvram.dat"
  End If
  
  FP = FreeFile
  Open App.Path & "\nvram.dat" For Binary Access Write As FP
  Put #FP, , NVRAM.RTC
  Put #FP, , NVRAM.Reg
  Put #FP, , NVRAM.RAM
  Close #FP
    
  ' we don't need to save the flags, these can be determined
  ' from the Reg values; and doing it like this will (at least)
  ' make the file 64 bytes that should match a dump of an actual
  ' NVRAM...
  
End Sub


Public Sub NVRAM_Write(ByVal Value As Byte, ByVal Addr As Integer)
  ' An attempt has been made to write to the NVRAM
  If (Latch.RTCOkay = False) Then
    MsgBox "Attempted to write to the NVRAM without having " & _
           "MFLatch bit 3 set...", vbExclamation + vbOKOnly
    Exit Sub
  End If
  
  If (Addr < 4) Then
    ' &FC00 (-&FC03) is mapped in as the 'address' selection
    NVRAMAddr = (Value And 63) ' is this sufficient?  ##TODO##
    Exit Sub
  End If
  
  ' &FC04 (-&FC07) is mapped in as the 'data' selection
  Select Case NVRAMAddr
    Case 0, 2, 4, 6, 7, 8, 9:
      ' RTC values, silently discard
    
    Case 1, 3, 5:
      ' Alarm, not currently supported!
      MsgBox "Attempted to set NVRAM alarm. This is not " & _
             "a currently-supported possibility." & vbCrLf & _
             "(please refer to ""NVRAMmod.bas"")", _
             vbExclamation + vbOKOnly
    
    Case 10:
      ' Internal register A
      NVRAM.Reg.RegA = (Value And 127) ' not the UIP bit
      NVRAM_RegDecode 10
      
    Case 11:
      ' Internal register B
      NVRAM.Reg.RegB = Value ' fully programmable
      NVRAM_RegDecode 11
      
    Case 12:
      ' Internal register C
      ' is not writeable
      
    Case 13:
      ' Internal register D
      ' is not writeable
      
    Case 14 To 63:
      ' NVRAM bytes
      NVRAM.RAM.RAM(NVRAMAddr) = Value
      
      If (NVRAM.RAM.RAM(NVRAMAddr) = 20) Then
        If (MsgBox("Error code &" & Hex(Value) & " was written to NVRAM from " & _
                   "code at &" & Right$("0000" & Hex$(CPU_ReadPC), 4) & "." & vbCrLf & _
                   "Would you like to stop execution at this point?", _
                   vbQuestion + vbYesNo) = vbYes) Then
          CPU_Abort
          ExecHalt = True
          frmCPUState.Show
          CPU_RefreshTrace ' only works when window is 'visible'
          frmCPUState.Hide
          frmCPUState.Show
        End If
      End If
  End Select
End Sub


Public Function NVRAM_Read(Addr As Integer) As Byte
  ' An attempt has been made to read from the NVRAM
  Dim Temp As Byte
  
  If (Latch.RTCOkay = False) Then
    MsgBox "Attempted to read from the NVRAM without having " & _
           "MFLatch bit 3 set...", vbExclamation + vbOKOnly
    Exit Function
  End If
  
  If (Addr < 4) Then
    MsgBox "Attempted to read address from NVRAM? Weird!", _
           vbExclamation + vbOKOnly
    ' it's latched into the device on the Address Strobe;
    ' so we can't actually perform this operation!
    Exit Function
  End If
  
  Select Case NVRAMAddr
    Case 0: ' Seconds
      NVRAM_Read = Second(Now)
    
    Case 2: ' Minutes
      NVRAM_Read = Minute(Now)
    
    Case 4: ' Hours
      Temp = CByte(Hour(Now))
      If (NVRAM.Flags.f2412 = False) Then
        ' 12 hour mode
        If (Temp > 12) Then
          Temp = Temp - 12
          Temp = Temp Or 128 ' high bit set means PM
        End If
      End If
      NVRAM_Read = Temp
    
    Case 6: ' Day of the Week
      NVRAM_Read = WeekDay(Now, vbSunday)
      ' given FileStore fiddles the date to start with 1981,
      ' this may be incorrect/unused in the 'real' hardware...
    
    Case 7: ' Day of the Month
      NVRAM_Read = Day(Now)
    
    Case 8: ' Month
      NVRAM_Read = Month(Now)
    
    Case 9: ' Year
      NVRAM_Read = Year(Now) - 1981 ' thus 2009 is 28...
    
    Case 1, 3, 5: ' Alarm
      MsgBox "Attempted to read NVRAM alarm. This is not " & _
             "a currently-supported possibility." & vbCrLf & _
             "(please refer to ""NVRAMmod.bas"")", _
             vbExclamation + vbOKOnly
    
    Case 10: ' Internal register A
      NVRAM_RegEncode
      NVRAM_Read = NVRAM.Reg.RegA

    Case 11: ' Internal register B
      NVRAM_RegEncode
      NVRAM_Read = NVRAM.Reg.RegB

    Case 12: ' Internal register C
      NVRAM_RegEncode
      NVRAM_Read = NVRAM.Reg.RegC
      ' has side effect of clearing IRQ flags
      ' poke it directly, it's quicker
      NVRAM.Flags.IRQF = False
      NVRAM.Flags.PF = False
      NVRAM.Flags.AF = False
      NVRAM.Flags.UF = False
      NVRAM.Reg.RegC = 0

    Case 13: ' Internal register D
      NVRAM_RegEncode
      NVRAM_Read = NVRAM.Reg.RegD
      ' has no side effects as we say VRT always set
      
    Case 14 To 63: ' RAM
      NVRAM_Read = NVRAM.RAM.RAM(NVRAMAddr)
  
      If (NVRAM.RAM.RAM(NVRAMAddr) = 49) Then
        If (MsgBox("Harddisc validity check option was read from NVRAM from " & _
                   "code at &" & Right$("0000" & Hex$(CPU_ReadPC), 4) & "." & vbCrLf & _
                   "Would you like to stop execution at this point?", _
                   vbQuestion + vbYesNo) = vbYes) Then
          CPU_Abort
          ExecHalt = True
          frmCPUState.Show
          CPU_RefreshTrace ' only works when window is 'visible'
          frmCPUState.Hide
          frmCPUState.Show
        End If
      End If
  End Select
End Function



Private Sub NVRAM_RegDecode(Optional ByVal Which As Integer = 0)
  ' Decodes the RegA-RegD bytes into the associated flags
  '
  ' The "Which" is used specifically when writing into the NVRAM
  ' in order to prevent warnings of BCD mode being set because
  ' we have just written register A and not yet touched register
  ' B (which is 0 so will appear as if BCD mode is active).
  '
  
  If ((Which = 10) Or (Which = 0)) Then
    ' RegisterA
    NVRAM.Flags.UIP = ((NVRAM.Reg.RegA And 128) = 128)
    NVRAM.Flags.DivDV = ((NVRAM.Reg.RegA \ 16) And 7)
    NVRAM.Flags.DivRS = (NVRAM.Reg.RegA And 15)
  
  ElseIf ((Which = 11) Or (Which = 0)) Then
    ' RegisterB
    NVRAM.Flags.fSET = ((NVRAM.Reg.RegB And 128) = 128)
    NVRAM.Flags.PIE = ((NVRAM.Reg.RegB And 64) = 64)
    NVRAM.Flags.AIE = ((NVRAM.Reg.RegB And 32) = 32)
    NVRAM.Flags.UIE = ((NVRAM.Reg.RegB And 16) = 16)
    NVRAM.Flags.SQWE = ((NVRAM.Reg.RegB And 8) = 8)
    NVRAM.Flags.DM = ((NVRAM.Reg.RegB And 4) = 4)
    NVRAM.Flags.f2412 = ((NVRAM.Reg.RegB And 2) = 2)
    NVRAM.Flags.DSE = ((NVRAM.Reg.RegB And 1) = 1)
    If (NVRAM.Flags.fSET = True) Then NVRAM.Flags.UIE = False
  
    ' BCD?
    If (NVRAM.Flags.DM = False) Then
      MsgBox "Attempted to set NVRAM RTC to BDC mode. This is " & _
             "not a currently-supported possibility." & vbCrLf & _
             "(please refer to ""NVRAMmod.bas"")", _
             vbExclamation + vbOKOnly
      ' will carry on in binary mode
    End If

    ' Interrupt on update completion?
    If (NVRAM.Flags.UIE = True) Then
      NVRAMCyc = Cycles + 2000000 ' next interrupt time
      If (NVRAMCyc > &H6FFFFFFF) Then NVRAMCyc = NVRAMCyc - &H6FFFFFFF
      ' trap disability wraparound
      If (NVRAMCyc = 0&) Then NVRAMCyc = 1&
    Else
      NVRAMCyc = 0
    End If

    ' ##TODO## Other interrupt vectors are not considered as
    '          they are not required by the FileStore...

  ElseIf ((Which = 12) Or (Which = 0)) Then
    ' RegisterC
    NVRAM.Flags.PF = ((NVRAM.Reg.RegC And 64) = 64)
    NVRAM.Flags.AF = ((NVRAM.Reg.RegC And 32) = 32)
    NVRAM.Flags.UF = ((NVRAM.Reg.RegC And 16) = 16)
    NVRAM.Flags.IRQF = (NVRAM.Reg.RegC > 0) ' set if any flag set

  ElseIf ((Which = 13) Or (Which = 0)) Then
    ' RegisterD - report ALWAYS valid data
    NVRAM.Flags.VRT = True
    NVRAM.Reg.RegD = 128
  End If
      
    
  ' The FileStore sets:
  ' NVRAM state:
  '   DivDV = 2        [32.768kHz clock]
  '   DivRS = 6        [976.562uS periodic interrupt]
  '     SET = False    Update the time once every second
  '     PIE = False    No periodic interrupt
  '     AIE = False    No alarm interrupt
  '     UIE = True     Notify of update completed (once/sec)
  '    SQWE = True     Output square wave on TP at 1.024kHz
  '      DM = True     Binary mode (not BCD)
  '    2412 = True     24hr clock
  '     DSE = False    No daylight savings (not supported by RTC!)

#If NVRAMREG = 1 Then
  Debug.Print "NVRAM state:"
  Debug.Print "  DivDV = " & CStr(NVRAM.Flags.DivDV)
  Debug.Print "  DivRS = " & CStr(NVRAM.Flags.DivRS)
  Debug.Print "    SET = " & CStr(NVRAM.Flags.fSET)
  Debug.Print "    PIE = " & CStr(NVRAM.Flags.PIE)
  Debug.Print "    AIE = " & CStr(NVRAM.Flags.AIE)
  Debug.Print "    UIE = " & CStr(NVRAM.Flags.UIE)
  Debug.Print "   SQWE = " & CStr(NVRAM.Flags.SQWE)
  Debug.Print "     DM = " & CStr(NVRAM.Flags.DM)
  Debug.Print "   2412 = " & CStr(NVRAM.Flags.f2412)
  Debug.Print "    DSE = " & CStr(NVRAM.Flags.DSE)
  Debug.Print ""
#End If
End Sub


Private Sub NVRAM_RegEncode()
  ' Converts flags into register values for RegA-RegD
  Dim Reg As Integer
  
  ' RegisterA
  Reg = (NVRAM.Reg.RegA And 127) ' mask off UIP
  If (NVRAM.Flags.UIP = True) Then Reg = Reg Or 128 ' add UIP?
  NVRAM.Reg.RegA = Reg
  
  ' RegisterB
  ' nothing here is updated by us, it's all user-config stuff
  
  ' RegisterC
  Reg = 0
  If (NVRAM.Flags.PF = True) Then Reg = Reg Or 64
  If (NVRAM.Flags.AF = True) Then Reg = Reg Or 32
  If (NVRAM.Flags.UF = True) Then Reg = Reg Or 16
  If (Reg > 0) Then
    ' set IRQ flag if any individual flags are set
    NVRAM.Flags.IRQF = True
    Reg = Reg Or 128
  End If
  NVRAM.Reg.RegC = Reg
  
  ' RegisterD - report ALWAYS valid data
  NVRAM.Flags.VRT = True
  NVRAM.Reg.RegD = 128
End Sub


Public Sub NVRAM_Poll()
  ' Called every 'x' instructions to see if we are ready to
  ' produce an interrupt. It isn't terribly accurate ATM, but
  ' it'll do for now.

  ' No ticker defined...
  If (NVRAMCyc = 0) Then Exit Sub
  
  ' SET bit enabled, no ticker activity...
  If (NVRAM.Flags.fSET = True) Then Exit Sub

  If (Cycles >= NVRAMCyc) Then
    NVRAM.Flags.UF = True
    NVRAM.Flags.IRQF = True
    NVRAM.Reg.RegC = 128 Or 16
    CPU_Vector IRQVEC
  
    NVRAMCyc = Cycles + 2000000 ' next interrupt time
    If (NVRAMCyc > &H6FFFFFFF) Then NVRAMCyc = NVRAMCyc - &H6FFFFFFF
    If (NVRAMCyc = 0&) Then NVRAMCyc = 1&
  End If
End Sub


Public Function NVRAM_ReadByte(ByVal Addr As Integer) As Byte
  If ((Addr < 14) Or (Addr > 63)) Then
    ' flag an error
    NVRAM_ReadByte = -1
  Else
    NVRAM_ReadByte = NVRAM.RAM.RAM(Addr)
  End If
End Function
